## Stand Alone Script for the B2B observation imaging                      
## (Customized for uid://A002/Xf0e992/X4d18, R Lep Band8 B2B)           
## Script to image the B2B B4 to B8
## R Lep testing data 
## NOTE this is not an SV dataset

## On round of selfcal is made using the continuum
## there are no lines detected in initial analysis
## and no image cubes are made. The Primary beam
## corrections are not applied. 

## Below 'EXTRA NOTES' sections provide some 
## more context of the steps 

### Hardcoded script Final Version
### Yoshiharu Asaki JAO & Luke Maud ESO
### For further information please contact:
### yasaki@alma.cl   JAO/NAOJ
### lmaud@eso.org   ESO
###
### If using these data please reference:
### Asaki et al 2023
### and B2B papers therein and published thereafter

### 13/02/2023 - CASA 6.4.1-12 version

### IMPORTANT
### SPWs 19,21,25,27 are the BAND 8 - science target frequency
### SPWs 7,35,39,41 are the BAND 4 - reference calibrator frequency
### these are in the image names 

thesteps = []
step_title = {0: 'Image the calibrators',
              1: 'Image Continuum R Lep',
              2: 'Selfcal and image R Lep',
              3: 'Copy out fits images'
             }

if 'applyonly' not in globals(): applyonly = False
try:
  print('List of steps to be executed ...'+str(mysteps))
  thesteps = mysteps
except:
  print('global variable mysteps not set.')
if (thesteps==[]):
  thesteps = range(0,len(step_title))
  print('Executing all steps: ', thesteps)

# The Python variable 'mysteps' will control which steps
# are executed when you start the script using
#   execfile('scriptForCalibration.py')
# e.g. setting
#   mysteps = [2,3,4]
# before starting the script will make the script execute
# only steps 2, 3, and 4
# Setting mysteps = [] will make it execute all steps.

import re

import os

import casalith

import glob

if re.search('^6.4.1', '.'.join([str(i) for i in casalith.version()[:-1]])) == None:
 sys.exit('ERROR: PLEASE USE THE SAME VERSION OF CASA THAT YOU USED FOR GENERATING THE SCRIPT: 6.4.1')

mystep = 0
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  ## EXTRA NOTE - mask for point source calibrators is fixed as 
  ##              a central circle. Image parameters are also
  ##              provided to make a smaller focused image
  ##              to reduce imaging time

  cell_size_im = '0.001arcsec'
  image_size = 512
  centre_mask = ['circle[[256pix,256pix],10pix]']
  nits = 100


  # Imaging the Bandpass/Flux cal - only HF

  os.system('rm -rf J0423-0120_bp.spw19_21_25_27.cont*')
  tclean(vis = 'uid___A002_Xf0e992_X4d18.ms.split.cal',
         imagename='J0423-0120_bp.spw19_21_25_27.cont', 
         field='J0423-0120',
         interactive=False,
         spw = '19,21,25,27',
         cell=cell_size_im, imsize=image_size, niter=nits,
         specmode='cont',
         weighting='briggs',
         robust=0.5,
         mask=centre_mask)

  # DGC at HF - B8 
  # EXTRA NOTE - recall the HF data at B8 are all amplitude
  #              calibrated w.r.t. the bandpass/flux cal

  os.system('rm -rf J0522-3627_dgc.spw19_21_25_27.cont*')
  tclean(vis = 'uid___A002_Xf0e992_X4d18.ms.split.cal',
         imagename='J0522-3627_dgc.spw19_21_25_27.cont', 
         field='J0522-3627',
         interactive=False,
         spw = '19,21,25,27',
         cell=cell_size_im, imsize=image_size, niter=nits,
         specmode='cont',
         weighting='briggs',
         robust=0.5,
         mask=centre_mask)


  # CHECK source at HF - B8 
  # EXTRA NOTE - recall the HF data at B8 are all amplitude
  #              calibrated w.r.t. the bandpass/flux cal.
  #              The check source is also located 'further'
  #              from the phase cal (~3deg) than the target
  #              thus Phase transfer will be reasonable, but
  #              not perfect. See also Maud et al 2023 

  os.system('rm -rf J0516-1603_chk.spw19_21_25_27.cont*')
  tclean(vis = 'uid___A002_Xf0e992_X4d18.ms.split.cal',
         imagename='J0516-1603_chk.spw19_21_25_27.cont', 
         field='J0516-1603',
         interactive=False,
         spw = '19,21,25,27',
         cell=cell_size_im, imsize=image_size, niter=nits,
         specmode='cont',
         weighting='briggs',
         robust=0.5,
         mask=centre_mask)


  # DGC at LF - B4 
  # EXTRA NOTE - recall the LF data at B4 are all amplitude
  #              calibrated w.r.t. the DGC 

  os.system('rm -rf J0522-3627_dgc.spw7_35_39_41.cont*')
  tclean(vis = 'uid___A002_Xf0e992_X4d18.ms.split.cal',
         imagename='J0522-3627_dgc.spw7_35_39_41.cont', 
         field='J0522-3627',
         interactive=False,
         spw = '7,35,39,41',
         cell=cell_size_im, imsize=image_size, niter=nits,
         specmode='cont',
         weighting='briggs',
         robust=0.5,
         mask=centre_mask)

  os.system('rm -rf J0504-1512_ph.spw7_35_39_41.cont*')
  tclean(vis = 'uid___A002_Xf0e992_X4d18.ms.split.cal',
         imagename='J0504-1512_ph.spw7_35_39_41.cont', 
         field='J0504-1512',
         interactive=False,
         spw = '7,35,39,41',
         cell=cell_size_im, imsize=image_size, niter=nits,
         specmode='cont',
         weighting='briggs',
         robust=0.5,
         mask=centre_mask)


mystep = 1
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  cell_size_im = '0.001arcsec'
  image_size = 512
  nits = 1000

  # Target continuum
  os.system('rm -rf R_Lep_sci.spw19_21_25_27.cont.I.manual.*')
  tclean(vis = 'uid___A002_Xf0e992_X4d18.ms.split.cal',
         imagename='R_Lep_sci.spw19_21_25_27.cont.I.manual',
         field='R_Lep',
         datacolumn = 'data',
         interactive=False,
         spw='19,21,25,27',         
         deconvolver='hogbom',
         phasecenter='ICRS 04h59m36.359000s -14d48m22.531000s',
         cell=cell_size_im, imsize=image_size, niter=nits,
         specmode='cont',
         weighting='briggs', 
         robust=0.5,
         usemask='auto-multithresh',
         sidelobethreshold=2.0,
         noisethreshold=3.00,
         lownoisethreshold=1.0,
         minbeamfrac=0.3,
         growiterations=75,
         negativethreshold=0.0,
         savemodel='modelcolumn')


mystep = 2
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  cell_size_im = '0.001arcsec'
  image_size = 512
  nits = 1000

  os.system('rm -rf R_Lep.p1*')
  gaincal(vis = 'uid___A002_Xf0e992_X4d18.ms.split.cal',
         caltable='R_Lep.p1',
         solint='inf',
         calmode='p',
         gaintype='T',
         spw = '19,21,25,27',
         combine = 'spw',
         minsnr=3.0,
         refant='DV02',
         field='R_Lep')

  applycal(vis = 'uid___A002_Xf0e992_X4d18.ms.split.cal',
        field = 'R_Lep',
        spw = '19,21,25,27',
        spwmap = [19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19,19],
        interp='linearPD',
        gaintable = 'R_Lep.p1',
        gainfield = '')

  os.system('rm -rf R_Lep_sci.spw19_21_25_27.cont.I.manual-selfcal*')
  tclean(vis = 'uid___A002_Xf0e992_X4d18.ms.split.cal',
         imagename='R_Lep_sci.spw19_21_25_27.cont.I.manual-selfcal',
         field='R_Lep',
         datacolumn= 'corrected',
         interactive=False,
         spw='19,21,25,27',         
         deconvolver='hogbom',
         phasecenter='ICRS 04h59m36.359000s -14d48m22.531000s',
         cell=cell_size_im, imsize=image_size, niter=nits,
         specmode='cont',
         weighting='briggs', 
         robust=0.5,
         usemask='auto-multithresh',
         sidelobethreshold=2.0,
         noisethreshold=3.00,
         lownoisethreshold=1.0,
         minbeamfrac=0.3,
         growiterations=75,
         negativethreshold=0.0,
         savemodel='modelcolumn')
       
mystep = 3
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  imagetofits=glob.glob('*.image*') 

  for imagefit in imagetofits: 
      exportfits(imagefit, fitsimage=imagefit+'.fits') 
