# ALMA Data Reduction Script

### This is a QA2-like script for calibration for the Band-to-Band
### testing dataset for R Lep observed at band 8 during the 2021 
### testing for the High-Frequency Long Baselines EOC campaign.
### Here, the B2B uses phase transfer from a phase calibrator 
### observed at band 4.
 
### Notes are also added 'EXTRA NOTES' during the steps below
### to help/inform inquisitive users. 

### These data were taken with SBs but specifically for
### testing purposes. These data were not, and are not science verification
### datasets and thus do not fully follow a 'standard' observational sequence.
### As the phase calibrator has some small structure steps are added
### to better correct small residual phase differences.

### For further information please contact:
### yasaki@alma.cl  JAO/NAOJ
### lmaud@eso.org   ESO 
###
### If using these data please reference:
### Asaki et al, 2023 ApJS, 
### and the B2B papers referenced therin

### The base scripts was produced by the alma script generator 
### owner and maintained by - Dirk Petry ESO
### Further modifications were made by the HF-LBC team
### Hardcoded script Final Version
### Yoshiharu Asaki JAO, Luke Maud ESO
### XX/02/2023 - CASA 6.4.1-12 version


thesteps = []
step_title = {0: 'Import of the ASDM',
              1: 'Fix of SYSCAL table times',
              2: 'listobs',
              3: 'A priori flagging',
              4: 'Generation of the WVR cal table',
              5: 'Generation of the Tsys cal table',
              6: 'Generation of the antenna position cal table',
              7: 'Application of the WVR, Tsys and antpos cal tables',
              8: 'Split out science SPWs',
              9: 'Listobs, and save original flags',
              10: 'Initial flagging',
              11: 'Putting a model for the flux calibrator(s)',
              12: 'Save flags before bandpass cal',
              13: 'Bandpass calibration',
              14: 'Save flags before gain cal',
              15: 'B2B Gain calibration Part I',
              16: 'B2B Gain calibration Part II',
              17: 'Save flags before applycal',
              18: 'Application of the bandpass and gain cal tables',
              19: 'Split out corrected column',
              20: 'Save flags after applycal'}

if 'applyonly' not in globals(): applyonly = False
try:
  print('List of steps to be executed ...'+str(mysteps))
  thesteps = mysteps
except:
  print('global variable mysteps not set.')
if (thesteps==[]):
  thesteps = range(0,len(step_title))
  print('Executing all steps: ', thesteps)

# The Python variable 'mysteps' will control which steps
# are executed when you start the script using
#   execfile('scriptForCalibration.py')
# e.g. setting
#   mysteps = [2,3,4]
# before starting the script will make the script execute
# only steps 2, 3, and 4
# Setting mysteps = [] will make it execute all steps.

import re

import os

import casalith

if applyonly != True: es = aU.stuffForScienceDataReduction() 


if re.search('^6.4.1', '.'.join([str(i) for i in casalith.version()[:-1]])) == None:
 sys.exit('ERROR: PLEASE USE THE SAME VERSION OF CASA THAT YOU USED FOR GENERATING THE SCRIPT: 6.4.1')


# CALIBRATE_AMPLI: 
#     field ids: ['']
# CALIBRATE_ATMOSPHERE: J0423-0120, J0522-3627, J0504-1512, R_Lep
#     field ids: [0, 1, 3, 5]
# CALIBRATE_BANDPASS: J0423-0120
#     field ids: [0]
# CALIBRATE_DIFFGAIN: J0522-3627
#     field ids: [1]

## EXTRA NOTES - for consistency with the B9 and B10 data
##               we do not use the targeted Flux calibrator, 
##               as this is testing data it was only observed
##               at the High Frequency. Instead we use
##               the bandpass source J0522-3627 which is 
##               targetted at the high and low frequency 
##               in all our observations

# CALIBRATE_FLUX: J0423-0120
#     field ids: [0]
# CALIBRATE_FOCUS: 
#     field ids: ['']
# CALIBRATE_PHASE: J0522-3627, J0504-1512
#     field ids: [1, 3]
# CALIBRATE_POINTING: J0423-0120, J0522-3627, J0438-1251
#     field ids: [0, 1, 2]
# CALIBRATE_POLARIZATION: 
#     field ids: ['']
# OBSERVE_CHECK: J0516-1603
#     field ids: [4]
# OBSERVE_TARGET: R_Lep
#     field ids: [5]


# Using reference antenna = DV02

# Import of the ASDM
mystep = 0
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  if os.path.exists('uid___A002_Xf0e992_X4d18.ms') == False:
    importasdm('uid___A002_Xf0e992_X4d18', asis='Antenna Station Receiver Source CalAtmosphere CalWVR CorrelatorMode SBSummary', bdfflags=True, lazy=False, process_caldevice=False)
    if not os.path.exists('uid___A002_Xf0e992_X4d18.ms.flagversions'):
      print('ERROR in importasdm. Output MS is probably not useful. Will stop here.')
      thesteps = []
  if applyonly != True: es.fixForCSV2555('uid___A002_Xf0e992_X4d18.ms')

# Fix of SYSCAL table times
mystep = 1
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  from casarecipes.almahelpers import fixsyscaltimes
  fixsyscaltimes(vis = 'uid___A002_Xf0e992_X4d18.ms')

print("# A priori calibration")

# listobs
mystep = 2
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  os.system('rm -rf uid___A002_Xf0e992_X4d18.ms.listobs')
  listobs(vis = 'uid___A002_Xf0e992_X4d18.ms',
    listfile = 'uid___A002_Xf0e992_X4d18.ms.listobs')
  
  

# A priori flagging
mystep = 3
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  flagdata(vis = 'uid___A002_Xf0e992_X4d18.ms',
    mode = 'manual',
    spw = '5~12,17~28,32~42',
    autocorr = True,
    flagbackup = False)
  
  flagdata(vis = 'uid___A002_Xf0e992_X4d18.ms',
    mode = 'manual',
    intent = '*POINTING*,*ATMOSPHERE*',
    flagbackup = False)
  
  flagcmd(vis = 'uid___A002_Xf0e992_X4d18.ms',
    inpmode = 'table',
    useapplied = True,
    action = 'plot',
    plotfile = 'uid___A002_Xf0e992_X4d18.ms.flagcmd.png')
  
  flagcmd(vis = 'uid___A002_Xf0e992_X4d18.ms',
    inpmode = 'table',
    useapplied = True,
    action = 'apply')
  

# Generation of the WVR cal table
mystep = 4
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  os.system('rm -rf uid___A002_Xf0e992_X4d18.ms.wvr') 
  
  os.system('rm -rf uid___A002_Xf0e992_X4d18.ms.wvrgcal') 
  
  # Warning: more than one integration time found on science data, I'm picking the lowest value. Please check this is right.
  
  mylogfile = casalog.logfile()
  casalog.setlogfile('uid___A002_Xf0e992_X4d18.ms.wvrgcal')
  
  wvrgcal(vis = 'uid___A002_Xf0e992_X4d18.ms',
    caltable = 'uid___A002_Xf0e992_X4d18.ms.wvr',
    spw = [7, 19, 21, 25, 27, 35, 39, 41],
    toffset = 0,
    tie = ['R_Lep,J0504-1512', 'J0516-1603,J0504-1512'],
    statsource = 'R_Lep')
  
  casalog.setlogfile(mylogfile)
  

  # EXTRA NOTES  - WVR is produced for all High and Low frequencies
  #                as part of B2B observations 


  if applyonly != True: aU.plotWVRSolutions(caltable='uid___A002_Xf0e992_X4d18.ms.wvr', spw='19', antenna='DV02',
    yrange=[-199,199],subplot=22, interactive=False,
    figfile='uid___A002_Xf0e992_X4d18.ms.wvr.plots/uid___A002_Xf0e992_X4d18.ms.wvrSignalSPWs') 
  
  if applyonly != True: aU.plotWVRSolutions(caltable='uid___A002_Xf0e992_X4d18.ms.wvr', spw='7', antenna='DV02',
    yrange=[-199,199],subplot=22, interactive=False,
    figfile='uid___A002_Xf0e992_X4d18.ms.wvr.plots/uid___A002_Xf0e992_X4d18.ms.wvrReferenceSPWs') 
  
  #Note: If you see wraps in these plots, try changing yrange or unwrap=True 
  #Note: If all plots look strange, it may be a bad WVR on the reference antenna.
  #      To check, you can set antenna='' to show all baselines.
  

# Generation of the Tsys cal table
mystep = 5
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  os.system('rm -rf uid___A002_Xf0e992_X4d18.ms.tsys') 
  gencal(vis = 'uid___A002_Xf0e992_X4d18.ms',
    caltable = 'uid___A002_Xf0e992_X4d18.ms.tsys',
    caltype = 'tsys')
  
  # Flagging edge channels
  
  flagdata(vis = 'uid___A002_Xf0e992_X4d18.ms.tsys',
    mode = 'manual',
    spw = '7:0~3;124~127,17:0~3;124~127,19:0~3;124~127,21:0~3;124~127,23:0~3;124~127,32:0~3;124~127,35:0~3;124~127,37:0~3;124~127',
    flagbackup = False)
  
  try: # protect against failure of plotbandpass in case of Tsys timestamp issues
    if applyonly != True: aU.plotbandpass(caltable='uid___A002_Xf0e992_X4d18.ms.tsys', overlay='time', 
      xaxis='freq', yaxis='amp', subplot=22, buildpdf=False, interactive=False,
      showatm=True,pwv='auto',chanrange='92.1875%',showfdm=True, showBasebandNumber=True, showimage=False, 
      field='', figfile='uid___A002_Xf0e992_X4d18.ms.tsys.plots.overlayTime/uid___A002_Xf0e992_X4d18.ms.tsys') 
  
  except:
    print('Error in plotbandpass. Skipping the time-overlay Tsys plots.')
  
  if applyonly != True: es.checkCalTable('uid___A002_Xf0e992_X4d18.ms.tsys', msName='uid___A002_Xf0e992_X4d18.ms', interactive=False) 
  

# Generation of the antenna position cal table
mystep = 6
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  # Warning: no baseline run found for following antenna(s): ['DA43', 'DA44', 'DA45', 'DA49', 'DA61', 'DV08', 'DV09', 'DV10', 'DV12', 'DV14', 'DV15', 'DV19', 'PM03'].
  
  os.system('rm -rf uid___A002_Xf0e992_X4d18.ms.antpos') 
  gencal(vis = 'uid___A002_Xf0e992_X4d18.ms',
    caltable = 'uid___A002_Xf0e992_X4d18.ms.antpos',
    caltype = 'antpos',
    antenna = 'DA42',
  #  parameter = [])
    parameter = [0,0,0])
  
  
  # antenna x_offset y_offset z_offset total_offset baseline_date
  

# Application of the WVR, Tsys and antpos cal tables
mystep = 7
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  # EXTRA NOTES  - Tsys is produced for all High and Low frequencies
  #                as part of B2B observations 

  
  from casarecipes.almahelpers import tsysspwmap
  tsysmap = tsysspwmap(vis = 'uid___A002_Xf0e992_X4d18.ms', tsystable = 'uid___A002_Xf0e992_X4d18.ms.tsys', tsysChanTol = 1)
  
  
  
  applycal(vis = 'uid___A002_Xf0e992_X4d18.ms',
    field = '0',
    spw = '19,21,25,27',
#    spw = '7,19,21,25,27',
    gaintable = ['uid___A002_Xf0e992_X4d18.ms.tsys', 'uid___A002_Xf0e992_X4d18.ms.wvr', 'uid___A002_Xf0e992_X4d18.ms.antpos'],
    gainfield = ['0', '', ''],
    interp = ['','nearest, linear',''],
    spwmap = [tsysmap,[],[]],
    calwt = True,
    flagbackup = False)
  
  
  
  applycal(vis = 'uid___A002_Xf0e992_X4d18.ms',
    field = '1',
    spw = '7,19,21,25,27,35,39,41',
    gaintable = ['uid___A002_Xf0e992_X4d18.ms.tsys', 'uid___A002_Xf0e992_X4d18.ms.wvr', 'uid___A002_Xf0e992_X4d18.ms.antpos'],
    gainfield = ['1', '', ''],
    interp = ['','nearest, linear',''],
    spwmap = [tsysmap,[],[]],
    calwt = True,
    flagbackup = False)
  
  
  
  # Note: J0438-1251 didn't have any Tsys measurement, and I couldn't find any close measurement. But this is not a science target, so this is probably Ok.
  
  applycal(vis = 'uid___A002_Xf0e992_X4d18.ms',
    field = '3',
    spw = '7,35,39,41',
    gaintable = ['uid___A002_Xf0e992_X4d18.ms.tsys', 'uid___A002_Xf0e992_X4d18.ms.wvr', 'uid___A002_Xf0e992_X4d18.ms.antpos'],
    gainfield = ['3', '', ''],
    interp = ['','nearest, linear',''],
    spwmap = [tsysmap,[],[]],
    calwt = True,
    flagbackup = False)
  
  
  
  # Note: J0516-1603 didn't have any Tsys measurement, so I used the one made on R_Lep. This is probably Ok.
  
  applycal(vis = 'uid___A002_Xf0e992_X4d18.ms',
    field = '4',
    spw = '19,21,25,27',
    gaintable = ['uid___A002_Xf0e992_X4d18.ms.tsys', 'uid___A002_Xf0e992_X4d18.ms.wvr', 'uid___A002_Xf0e992_X4d18.ms.antpos'],
    gainfield = ['5', '', ''],
    interp = ['','nearest, linear',''],
    spwmap = [tsysmap,[],[]],
    calwt = True,
    flagbackup = False)
  
  
  
  applycal(vis = 'uid___A002_Xf0e992_X4d18.ms',
    field = '5',
    spw = '19,21,25,27',
    gaintable = ['uid___A002_Xf0e992_X4d18.ms.tsys', 'uid___A002_Xf0e992_X4d18.ms.wvr', 'uid___A002_Xf0e992_X4d18.ms.antpos'],
    gainfield = ['5', '', ''],
    interp = ['','nearest, linear',''],
    spwmap = [tsysmap,[],[]],
    calwt = True,
    flagbackup = False)
  
  
  
  if applyonly != True: es.getCalWeightStats('uid___A002_Xf0e992_X4d18.ms', 0.2)
  

# Split out science SPWs
mystep = 8
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  os.system('rm -rf uid___A002_Xf0e992_X4d18.ms.split') 
  os.system('rm -rf uid___A002_Xf0e992_X4d18.ms.split.flagversions') 

  # EXTRA NOTES  - the mstransform splits out the High frequency 
  #                science target SPWs and the low frequency 
  #                phase calibator SPWs
  
  

  mstransform(vis = 'uid___A002_Xf0e992_X4d18.ms',
    outputvis = 'uid___A002_Xf0e992_X4d18.ms.split',
    datacolumn = 'corrected',
    spw = '7,19,21,25,27,35,39,41',
    reindex = False,
    keepflags = True)


  
  

print("# Calibration")

# Listobs, and save original flags
mystep = 9
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  os.system('rm -rf uid___A002_Xf0e992_X4d18.ms.split.listobs')
  listobs(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    listfile = 'uid___A002_Xf0e992_X4d18.ms.split.listobs')
  
  
  if not os.path.exists('uid___A002_Xf0e992_X4d18.ms.split.flagversions/Original.flags'):
    flagmanager(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
      mode = 'save',
      versionname = 'Original')
  
  

# Initial flagging
mystep = 10
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  # Flagging shadowed data
  
  flagdata(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    mode = 'shadow',
    flagbackup = False)
  
  # Flagging edge channels
  
  flagdata(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    mode = 'manual',
    spw = '7:0~7;120~127,19:0~7;120~127,21:0~7;120~127,35:0~7;120~127',
    flagbackup = False)
  
  

# Putting a model for the flux calibrator(s)
mystep = 11
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])


  ##  EXTRA NOTES - we use the source J0522-3627
  ##                for flux/amplitude calibration 
  ##                as it is present in all our 
  ##                observations for R Lep
  ##                This field is the DGC in these data


  setjy(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    standard = 'manual',
    field = 'J0522-3627',
    spw = '7,19,21,25,27,35,39,41',
    usescratch = True,
    fluxdensity = [4.480897994672396, 0, 0, 0],
    spix = -0.33113955903252534,
    reffreq = '405.0302985542997GHz')
  
  # fluxDensityUncertainty = 0.09569794381594136
  # meanAge = 2.0
  

# Save flags before bandpass cal
mystep = 12
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  
  flagmanager(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    mode = 'save',
    versionname = 'BeforeBandpassCalibration')
  
  

# Bandpass calibration
mystep = 13
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])


  # EXTRA NOTES - due to the arrangement of these testing B2B data
  #               there is no specific designated BANDPASS source
  #               for the Low Frequency. In this case, we can use
  #               the DGC source as it is a strong point source 
  #          
  #               In the "ap_pre" plots the LF is shown in 4 groups
  #               which form the DGC scan blocks


  os.system('rm -rf uid___A002_Xf0e992_X4d18.ms.split.ap_pre_bandpass') 
  
  gaincal(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    caltable = 'uid___A002_Xf0e992_X4d18.ms.split.ap_pre_bandpass',
    field = '0', # J0423-0120
    spw = '19:0~127,21:0~127,25:0~1919,27:0~1919',
    scan = '3',
    solint = 'int',
    refant = 'DV02',
    refantmode = 'strict',
    minsnr = 3.0,
    calmode = 'p')
  
  # EXTRA NOTE - for the LF all the scans of the DGC are used

  gaincal(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    caltable = 'uid___A002_Xf0e992_X4d18.ms.split.ap_pre_bandpass',
    field = '1', # J0522-3627
    spw = '7:0~127,19:0~127,21:0~127,25:0~1919,27:0~1919,35:0~127,39:0~1919,41:0~1919',
    scan = '7,9,11,13,15,17,71,73,75,77,79,81,136,138,140,142,144,146,199,201,203,205,207,209',
    solint = 'int',
    minsnr = 3.0,
    refant = 'DV02',
    refantmode = 'strict',
    calmode = 'p',
    append = True)
  
  if applyonly != True: es.checkCalTable('uid___A002_Xf0e992_X4d18.ms.split.ap_pre_bandpass', msName='uid___A002_Xf0e992_X4d18.ms.split', interactive=False) 
  
  os.system('rm -rf uid___A002_Xf0e992_X4d18.ms.split.bandpass') 
  bandpass(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    caltable = 'uid___A002_Xf0e992_X4d18.ms.split.bandpass',
    field = '0', # J0423-0120
    scan = '3',
    solint = 'inf',
    minsnr = 3.0,
    combine = 'scan',
    refant = 'DV02',
    solnorm = True,
    bandtype = 'B',
    gaintable = 'uid___A002_Xf0e992_X4d18.ms.split.ap_pre_bandpass')

  # EXTRA NOTE - for the LF all the scans of the DGC are used

  bandpass(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    caltable = 'uid___A002_Xf0e992_X4d18.ms.split.bandpass',
    field = '1', # J0522-3627
    scan = '7,9,11,13,15,17,71,73,75,77,79,81,136,138,140,142,144,146,199,201,203,205,207,209',
    solint = 'inf',
    minsnr = 3.0,
    combine = 'scan',
    refant = 'DV02',
    solnorm = True,
    bandtype = 'B',
    gaintable = 'uid___A002_Xf0e992_X4d18.ms.split.ap_pre_bandpass',
    append = True)
  
  if applyonly != True: es.checkCalTable('uid___A002_Xf0e992_X4d18.ms.split.bandpass', msName='uid___A002_Xf0e992_X4d18.ms.split', interactive=False) 
  

# Save flags before gain cal
mystep = 14
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  
  flagmanager(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    mode = 'save',
    versionname = 'BeforeGainCalibration')
  
  

# B2B Gain calibration Part I
mystep = 15
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  # do a phase offset for LF so all SPW can be merged for better SNR
  # if it is later required 

  os.system('rm -rf uid___A002_Xf0e992_X4d18.ms.split.DGCphaselow_phasediff')
  gaincal(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    caltable = 'uid___A002_Xf0e992_X4d18.ms.split.DGCphaselow_phasediff',
    field = '1',
    scan = '7,9,11,13,15,17,71,73,75,77,79,81,136,138,140,142,144,146,199,201,203,205,207,209',
    spw = '7,35,39,41',
    solint = 'inf',
    refant = 'DV02',
    gaintype = 'G',
    calmode = 'p',
    combine = 'scan',
    minsnr = 3.0,
    gaintable = 'uid___A002_Xf0e992_X4d18.ms.split.bandpass')
  
  if applyonly != True: es.checkCalTable('uid___A002_Xf0e992_X4d18.ms.split.DGCphaselow_phasediff', msName='uid___A002_Xf0e992_X4d18.ms.split', interactive=False)
  
  # LF temporal solutions fast

  # EXTRA NOTE - we DO NOT need to combine the SPW as the SNR is high
  #              but such a process is possible. One would need 
  #              to specific the combine='spw', but also later
  #              be very careful to ALWAYS use a correct spwmap
  #              list when appliying the calibrations
  #
  #             refantmode = strict is always used as we need to
  #             keep it fixed for B2B cross SPW calibration

  os.system('rm -rf uid___A002_Xf0e992_X4d18.ms.split.DGCphaselow_int')
  gaincal(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    caltable = 'uid___A002_Xf0e992_X4d18.ms.split.DGCphaselow_int',
    field = '1',
    scan = '7,9,11,13,15,17,71,73,75,77,79,81,136,138,140,142,144,146,199,201,203,205,207,209',
    spw = '7,35,39,41',
    solint = 'int',
    refant = 'DV02',
    gaintype = 'G',
    calmode = 'p',
    minsnr = 3.0,
    refantmode = 'strict',
    interp = ['linear','linear'],
    gaintable = ['uid___A002_Xf0e992_X4d18.ms.split.bandpass', 'uid___A002_Xf0e992_X4d18.ms.split.DGCphaselow_phasediff'],
    gainfield = ['',''])
  
  if applyonly != True: es.checkCalTable('uid___A002_Xf0e992_X4d18.ms.split.DGCphaselow_int', msName='uid___A002_Xf0e992_X4d18.ms.split', interactive=False)
  
  # LF temporal solutions slow
  os.system('rm -rf uid___A002_Xf0e992_X4d18.ms.split.DGCphaselow_inf')
  gaincal(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    caltable = 'uid___A002_Xf0e992_X4d18.ms.split.DGCphaselow_inf',
    field = '1',
    scan = '7,9,11,13,15,17,71,73,75,77,79,81,136,138,140,142,144,146,199,201,203,205,207,209',
    spw = '7,35,39,41',
    solint = 'inf',
    refant = 'DV02',
    gaintype = 'G',
    calmode = 'p',
    minsnr = 3.0,
    refantmode = 'strict',
    interp = ['linear','linear'],
    gaintable = ['uid___A002_Xf0e992_X4d18.ms.split.bandpass','uid___A002_Xf0e992_X4d18.ms.split.DGCphaselow_phasediff'],
    gainfield = ['',''])
  
  if applyonly != True: es.checkCalTable('uid___A002_Xf0e992_X4d18.ms.split.DGCphaselow_inf', msName='uid___A002_Xf0e992_X4d18.ms.split', interactive=False)
  
  ## HF fast solutions
  os.system('rm -rf uid___A002_Xf0e992_X4d18.ms.split.DGCphasehigh_int')
  gaincal(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    caltable = 'uid___A002_Xf0e992_X4d18.ms.split.DGCphasehigh_int',
    field = '1',
    scan= '8,10,12,14,16,72,74,76,78,80,137,139,141,143,145,200,202,204,206,208',
    spw = '19,21,25,27',
    solint = 'int',
    refant = 'DV02',
    gaintype = 'G',
    calmode = 'p',
    minsnr = 3.0,
    refantmode = 'strict',
    interp = ['linear','linear'],
    gaintable = 'uid___A002_Xf0e992_X4d18.ms.split.bandpass',
    gainfield = [''])
  
  if applyonly != True: es.checkCalTable('uid___A002_Xf0e992_X4d18.ms.split.DGCphasehigh_int', msName='uid___A002_Xf0e992_X4d18.ms.split', interactive=False)
  
  # get the DGC band-offset
  # Here we apply the LF to the HF and find the offset for each DGC 'group'
  # if there are multiple DGC sequences - this does each one independently 

  # EXTRA NOTES - these plots should show very similar phase solutions
  #              for each of the 4 times. Some data are flagged and have
  #              less than 4 solutions (not an issue)
  #              take care not to over interpret when the phase wraps past
  #              +/-180 deg this is only how the plot is shown, not a phase change

  os.system('rm -rf uid___A002_Xf0e992_X4d18.ms.split.DGCoffset_multi')
  gaincal(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    caltable = 'uid___A002_Xf0e992_X4d18.ms.split.DGCoffset_multi',
    field = '1',
    scan = '8,10,12,14,16,72,74,76,78,80,137,139,141,143,145,200,202,204,206,208',
    spw = '19,21,25,27',
    solint = '8min',  # max. duration of one DGC visit block
    refant = 'DV02',
    refantmode = 'strict',
    gaintype = 'G',
    calmode = 'p',
    minsnr=3.0,
    combine='scan',
    interp=['linear','linearPD'],
    spwmap=[[],[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 7, 20, 35, 22, 23, 24, 39, 26, 41]],
    gainfield = ['',''],
    gaintable = ['uid___A002_Xf0e992_X4d18.ms.split.bandpass','uid___A002_Xf0e992_X4d18.ms.split.DGCphaselow_inf'])
  
  if applyonly != True: es.checkCalTable('uid___A002_Xf0e992_X4d18.ms.split.DGCoffset_multi', msName='uid___A002_Xf0e992_X4d18.ms.split', interactive=False)
  
  # EXTRA NOTE - the below will only be made specifically if data
  #              reduction is performed step by step through 
  #              the calibration, i.e. applyonly = False
  #              if applyonly is globally set to True this is not done
  
  #              The spread should be usually within +/-50 deg for well
  #              calibrated data. We note DV04 has slightly larger spread 
  #              in the last scan, but the overall DGC 'offset' plot
  #              show the B2B offset is consistent with the other solutions

  if applyonly != True:
    # QA2 check to verify the slow (inf) LF to HF and offset worked
    # apply the LF to the HF and the multi offset
    os.system('rm -rf uid___A002_Xf0e992_X4d18.ms.split.DGCresidual_offset_multi')
    gaincal(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
      caltable = 'uid___A002_Xf0e992_X4d18.ms.split.DGCresidual_offset_multi',
      field = '1',
      scan = '8,10,12,14,16,72,74,76,78,80,137,139,141,143,145,200,202,204,206,208',
      spw = '19,21,25,27',
      solint = 'inf',
      refant = 'DV02',
      refantmode = 'strict',
      gaintype = 'G',
      calmode = 'p',
      minsnr=3.0,
      interp=['linear','linearPD','linear'],
      spwmap=[[],[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 7, 20, 35, 22, 23, 24, 39, 26, 41],[]],
      gainfield = ['','',''],
      gaintable = ['uid___A002_Xf0e992_X4d18.ms.split.bandpass','uid___A002_Xf0e992_X4d18.ms.split.DGCphaselow_inf','uid___A002_Xf0e992_X4d18.ms.split.DGCoffset_multi'])
  
  if applyonly != True: es.checkCalTable('uid___A002_Xf0e992_X4d18.ms.split.DGCresidual_offset_multi', msName='uid___A002_Xf0e992_X4d18.ms.split', interactive=False)
  

  # EXTRA NOTE - the below will only be made specifically if data
  #              reduction is performed step by step through 
  #              the calibration, i.e. applyonly = False
  #              if appluyonly is globally set to True this is not done

  if applyonly != True:
    # QA2 check to verify the fast (int) LF to HF and offset worked
    os.system('rm -rf uid___A002_Xf0e992_X4d18.ms.split.DGCresidual_offset_multi_int')
    gaincal(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
      caltable = 'uid___A002_Xf0e992_X4d18.ms.split.DGCresidual_offset_multi_int',
      field = '1',
      scan = '8,10,12,14,16,72,74,76,78,80,137,139,141,143,145,200,202,204,206,208',
      spw = '19,21,25,27',
      solint = 'int',
      refant = 'DV02',
      refantmode = 'strict',
      gaintype = 'G',
      calmode = 'p',
      minsnr=3.0,
      interp=['linear','linearPD','linear'],
      spwmap=[[],[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 7, 20, 35, 22, 23, 24, 39, 26, 41],[]],
      gainfield = ['','',''],
      gaintable = ['uid___A002_Xf0e992_X4d18.ms.split.bandpass','uid___A002_Xf0e992_X4d18.ms.split.DGCphaselow_inf','uid___A002_Xf0e992_X4d18.ms.split.DGCoffset_multi'])
  
  if applyonly != True: es.checkCalTable('uid___A002_Xf0e992_X4d18.ms.split.DGCresidual_offset_multi_int', msName='uid___A002_Xf0e992_X4d18.ms.split', interactive=False)
  

# B2B Gain calibration Part II
mystep = 16
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  # LF fast phase as a QA2 check for stability scan drop-outs
  os.system('rm -rf uid___A002_Xf0e992_X4d18.ms.split.phaselow_int')
  gaincal(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    caltable = 'uid___A002_Xf0e992_X4d18.ms.split.phaselow_int',
    field = '1,3',
    spw = '7,35,39,41',
    solint = 'int',
    refant = 'DV02',
    refantmode = 'strict',
    gaintype = 'G',
    calmode = 'p',
    minsnr = 3.0,
    gaintable = ['uid___A002_Xf0e992_X4d18.ms.split.bandpass','uid___A002_Xf0e992_X4d18.ms.split.DGCphaselow_phasediff'])
  
  if applyonly != True: es.checkCalTable('uid___A002_Xf0e992_X4d18.ms.split.phaselow_int', msName='uid___A002_Xf0e992_X4d18.ms.split', interactive=False)
  
  # LF slow phase solution
  os.system('rm -rf uid___A002_Xf0e992_X4d18.ms.split.phaselow_inf')
  gaincal(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    caltable = 'uid___A002_Xf0e992_X4d18.ms.split.phaselow_inf',
    field = '1,3',
    spw = '7,35,39,41',
    solint = 'inf',
    refant = 'DV02',
    refantmode='strict',
    gaintype = 'G',
    calmode = 'p',
    minsnr = 3.0,
    gaintable = ['uid___A002_Xf0e992_X4d18.ms.split.bandpass','uid___A002_Xf0e992_X4d18.ms.split.DGCphaselow_phasediff'])
  
  if applyonly != True: es.checkCalTable('uid___A002_Xf0e992_X4d18.ms.split.phaselow_inf', msName='uid___A002_Xf0e992_X4d18.ms.split', interactive=False)
  


  ## HF phase up on all strong sources
  ## i.e. the BP, and DGC for amp gain table
  os.system('rm -rf uid___A002_Xf0e992_X4d18.ms.split.phasehigh_int')
  gaincal(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    caltable = 'uid___A002_Xf0e992_X4d18.ms.split.phasehigh_int',
    field = '0,1',
    spw = '19,21,25,27',
    solint = 'int',
    refant = 'DV02',
    refantmode = 'strict',
    gaintype = 'G',
    calmode = 'p',
    minsnr = 3.0,
    gaintable = 'uid___A002_Xf0e992_X4d18.ms.split.bandpass')
  
  if applyonly != True: es.checkCalTable('uid___A002_Xf0e992_X4d18.ms.split.phasehigh_int', msName='uid___A002_Xf0e992_X4d18.ms.split', interactive=False)
  

  # EXTRA NOTE - here we calibrate the amplitudes 
  #              we are scaling from field 1, the DGC
  #              J0522-3627 in this case as it is 
  #              present at both frequencies and in
  #              all our datasets

  # Amplitude calibration HF
  os.system('rm -rf uid___A002_Xf0e992_X4d18.ms.split.ampli_inf')
  gaincal(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    caltable = 'uid___A002_Xf0e992_X4d18.ms.split.ampli_inf',
    field = '1',
    scan = '8,10,12,14,16,72,74,76,78,80,137,139,141,143,145,200,202,204,206,208',
    spw = '19,21,25,27',
    solint = '5min',  # for each DGC block to be separate
    combine = 'scan',
    refant = 'DV02',
    gaintype = 'T',
    calmode = 'a',
    minsnr = 3.0,
    gaintable = ['uid___A002_Xf0e992_X4d18.ms.split.bandpass','uid___A002_Xf0e992_X4d18.ms.split.phasehigh_int'],
    gainfield = ['0',''])
  
  # Amplitude calibration LF, append
  gaincal(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    caltable = 'uid___A002_Xf0e992_X4d18.ms.split.ampli_inf',
    field = '1',
    spw = '7,35,39,41',
    solint = '5min', # each DGC block in time
    combine = 'scan',
    refant = 'DV02',
    gaintype = 'T',
    calmode = 'a',
    append = True,
    minsnr = 3.0,
    gaintable = ['uid___A002_Xf0e992_X4d18.ms.split.bandpass','uid___A002_Xf0e992_X4d18.ms.split.DGCphaselow_phasediff','uid___A002_Xf0e992_X4d18.ms.split.phaselow_int'],
    gainfield = ['1','',''])
  
  if applyonly != True: es.checkCalTable('uid___A002_Xf0e992_X4d18.ms.split.ampli_inf', msName='uid___A002_Xf0e992_X4d18.ms.split', interactive=False)

  

# Save flags before applycal
mystep = 17
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  
  flagmanager(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    mode = 'save',
    versionname = 'BeforeApplycal')
  
  

# Application of the bandpass and gain cal tables
mystep = 18
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  # calibrating the bandpass
  # EXTRA NOTE - amplitude scaling comes from DGC field 1
  applycal(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    field = '0',
    spw = '19,21,25,27',
    scan = '3',
    interp=['linear','linear','linear'],
    gaintable = ['uid___A002_Xf0e992_X4d18.ms.split.bandpass','uid___A002_Xf0e992_X4d18.ms.split.phasehigh_int','uid___A002_Xf0e992_X4d18.ms.split.ampli_inf'],
    gainfield = ['0', '0', '1'])
  
  # calibrating the DGC with the HF phases
  # this uses B2B transfer from the LOW FREQUENCY 

  # EXTRA NOTE - a rough guide on the B2B accuracy can be to compare the image 
  #              flux from the DGC HF image as calibrated with a B2B transfer from
  #              the low frequency, with the numbers from self-cal in the 
  #              fluxscale file

  applycal(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    field = '1',
    spw = '19,21,25,27',
    scan = '8,10,12,14,16,72,74,76,78,80,137,139,141,143,145,200,202,204,206,208',
    interp=['linear','linearPD','linear','linear'],
    gaintable = ['uid___A002_Xf0e992_X4d18.ms.split.bandpass','uid___A002_Xf0e992_X4d18.ms.split.phaselow_inf','uid___A002_Xf0e992_X4d18.ms.split.DGCoffset_multi','uid___A002_Xf0e992_X4d18.ms.split.ampli_inf'],
    spwmap = [[],[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 7, 20, 35, 22, 23, 24, 39, 26, 41],[],[]],
    gainfield = ['0','1','1','1'])
  
  ## calibrating the target
  # EXTRA NOTE - amplitude scaling comes from DGC field 1
  applycal(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    field = '5',
    spw = '19,21,25,27',
    interp = ['linear','linearPD', 'linear','linear'],
    gaintable = ['uid___A002_Xf0e992_X4d18.ms.split.bandpass','uid___A002_Xf0e992_X4d18.ms.split.phaselow_inf','uid___A002_Xf0e992_X4d18.ms.split.DGCoffset_multi','uid___A002_Xf0e992_X4d18.ms.split.ampli_inf'],
    spwmap = [[],[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 7, 20, 35, 22, 23, 24, 39, 26, 41],[],[]],
    gainfield = ['0', '3','1','1'])
  
  ## calibrating the checksource
  applycal(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    field = '4',
    spw = '19,21,25,27',
    interp = ['linear','linearPD','linear','linear'],
    gaintable = ['uid___A002_Xf0e992_X4d18.ms.split.bandpass','uid___A002_Xf0e992_X4d18.ms.split.phaselow_inf','uid___A002_Xf0e992_X4d18.ms.split.DGCoffset_multi','uid___A002_Xf0e992_X4d18.ms.split.ampli_inf'],
    spwmap = [[],[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 7, 20, 35, 22, 23, 24, 39, 26, 41],[],[]],
    gainfield = ['0','3','1','1'])
  


  ## calibration of the LF data
  # EXTRA NOTE - amplitude scaling comes from DGC field 1
  
  # calibrating the DGC LF
  applycal(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    field = '1',
    spw = '7,35,39,41',
    scan = '7,9,11,13,15,17,71,73,75,77,79,81,136,138,140,142,144,146,199,201,203,205,207,209',
    interp=['linear','linear','linear','linear'],
    gaintable = ['uid___A002_Xf0e992_X4d18.ms.split.bandpass','uid___A002_Xf0e992_X4d18.ms.split.DGCphaselow_phasediff','uid___A002_Xf0e992_X4d18.ms.split.phaselow_inf','uid___A002_Xf0e992_X4d18.ms.split.ampli_inf'],
    gainfield = ['1','','1','1'])
  
  # calibrating the phasecal LF
  applycal(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    field = '3',
    spw = '7,35,39,41',
    interp=['linear','linear','linear','linear'],
    gaintable = ['uid___A002_Xf0e992_X4d18.ms.split.bandpass','uid___A002_Xf0e992_X4d18.ms.split.DGCphaselow_phasediff','uid___A002_Xf0e992_X4d18.ms.split.phaselow_inf','uid___A002_Xf0e992_X4d18.ms.split.ampli_inf'],
    gainfield = ['1','','3','1'])
  

# Split out corrected column
mystep = 19
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  listOfIntents = ['CALIBRATE_BANDPASS#ON_SOURCE',
   'CALIBRATE_DIFFGAIN#REFERENCE',
   'CALIBRATE_DIFFGAIN#SIGNAL',
   'CALIBRATE_FLUX#ON_SOURCE',
   'CALIBRATE_PHASE#ON_SOURCE',
   'CALIBRATE_PHASE#REFERENCE',
   'CALIBRATE_WVR#AMBIENT',
   'CALIBRATE_WVR#HOT',
   'CALIBRATE_WVR#OFF_SOURCE',
   'CALIBRATE_WVR#ON_SOURCE',
   'CALIBRATE_WVR#REFERENCE',
   'CALIBRATE_WVR#SIGNAL',
   'OBSERVE_CHECK_SOURCE#ON_SOURCE',
   'OBSERVE_TARGET#ON_SOURCE']
  
  mstransform(vis = 'uid___A002_Xf0e992_X4d18.ms.split',
    outputvis = 'uid___A002_Xf0e992_X4d18.ms.split.cal',
    datacolumn = 'corrected',
    intent = ','.join(listOfIntents),
    keepflags = True)
  

  ## EXTRA NOTE - Target science SPW are 19,21,25,27 

# Save flags after applycal
mystep = 20
if(mystep in thesteps):
  casalog.post('Step '+str(mystep)+' '+step_title[mystep],'INFO')
  print('\nStep '+str(mystep)+' '+step_title[mystep])

  
  flagmanager(vis = 'uid___A002_Xf0e992_X4d18.ms.split.cal',
    mode = 'save',
    versionname = 'AfterApplycal')
  
  
